from ast import Str
from typing_extensions import override
from masterpiece import Plugin, Composite
from .openweathermap import OpenWeatherMap


class OpenWeatherMapPlugin(Plugin):
    """Plugin class, for installing a OpenWeatherMap classes into the host application."""

    def __init__(self, name: str = "openweather_map") -> None:
        """Create systemstatus object."""
        super().__init__(name)

    @override
    def install(self, app: Composite) -> None:
        # Create and insert a OpenWeathermap object into the host application.
        obj = OpenWeatherMap()
        app.add(obj)



"""
Description
===========

OpenWeatherMap weather forecast.

"""

from .openweathermap import OpenWeatherMap
from .openweathermap import OpenWeatherMapThread
from .openweathermap_plugin import OpenWeatherMapPlugin

__all__ = [
    "OpenWeatherMap",
    "OpenWeatherMapThread",
    "OpenWeatherMapPlugin"
]



import json
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, cast
from typing_extensions import override

from masterpiece.mqtt import MqttMsg, Mqtt
from juham_core import JuhamCloudThread, JuhamThread


class OpenWeatherMapThread(JuhamCloudThread):
    """Asynchronous thread for acquiring forecast from the open weathermap.org web service
    site.
    Note: the service does not include solar power cast."""

    _forecast_topic = ""
    _interval: float = 60
    _appid = "f3954a78a9eaa74096305a29054d8c88"  # obtain from openweathermap.org
    _location = "Lahti,fi"
    _url = ""

    def __init__(self, client: Optional[Mqtt] = None) -> None:
        """Construct with the given mqtt client. Acquires data from the visual
        crossing web service and publishes the forecast data to
        ```forecast_topic```.

        Args:
            client (object, optional): MQTT client. Defaults to None.
        """
        super().__init__(client)
        self.mqtt_client: Optional[Mqtt] = client

    def init(
        self,
        forecast_topic: str,
        location: str,
        appid: str,
        url: str,
        update_interval: float,
    ):
        """Initialize thread

        Args:
            forecast_topic (str): topic to post the forecast
            location (str): geographic location
            appid (str): appid for the openweathermap service
            url (str): url to openweathermap service
            update_interval (float): update interval
        """
        self._forecast_topic = forecast_topic
        self._location = location
        self._appid = appid
        self._url = url
        self._interval = update_interval

    @override
    def update_interval(self) -> float:
        return self._interval

    @override
    def make_weburl(self) -> str:
        return f"{self._url}?q={self._location}&APPID={self._appid}"

    @override
    def process_data(self, data: Any) -> None:
        super().process_data(data)
        r = data.json()

        forecast: List = []

        for r in r["list"]:
            dt: float = r["dt"]
            # dt_txt: str = r["dt_txt"]
            temp: float = r["main"]["temp"]
            feels: float = r["main"]["feels_like"]
            pressure: float = r["main"]["pressure"]
            humidity: float = r["main"]["humidity"]
            windspeed: float = r["wind"]["speed"]

            rec = {
                "id": "openweathermap",
                "ts": dt,
                "hour": datetime.fromtimestamp(dt, tz=timezone.utc).strftime("%H"),
                "day": datetime.fromtimestamp(dt, tz=timezone.utc).strftime("%Y%m%d%H"),
                "temp": temp - 273,  # from Kelvin to Celcius
                "feels": feels - 273,
                "pressure": pressure,
                "humidity": humidity,
                "windspeed": windspeed,
            }
            forecast.append(rec)

        msg = json.dumps(forecast)
        self.publish(self._forecast_topic, msg, qos=1, retain=True)
        self.info(
            f"OpenWeatherMap forecast for the next {len(forecast)} days published"
        )


class OpenWeatherMap(JuhamThread):
    """This class constructs a data acquisition object for reading weather
    forecasts from 'openweathermap.org'. It subscribes to the
    forecast topic and writes hourly data such as solar energy, temperature,
    and other attributes relevant to home automation into a time series
    database.

    Spawns an asynchronous thread to run queries at the specified
    update_interval.
    """

    _OPENWEATHERMAP: str = "OpenWeatherMap"
    workerThreadId: str = OpenWeatherMapThread.get_class_id()
    bucket: str = "OpenWeather"
    appid: str = "f3954a78a9eaa74096305a29054d8c88"
    location: str = "Lahti,fi"
    url: str = "https://api.openweathermap.org/data/2.5/forecast"
    update_interval: float = 12 * 3600

    def __init__(self, name: str = "openweathermap") -> None:
        super().__init__(name)
        self.worker: Optional[OpenWeatherMapThread] = None
        self.forecast_topic: str = self.make_topic_name("forecast")

    @override
    def on_connect(self, client: object, userdata: Any, flags: int, rc: int) -> None:
        super().on_connect(client, userdata, flags, rc)
        if rc == 0:
            self.subscribe(self.forecast_topic)

    @override
    def on_message(self, client: object, userdata: Any, msg: MqttMsg) -> None:
        if msg.topic == self.forecast_topic:
            em = json.loads(msg.payload.decode())
            self.on_forecast(em)
        else:
            super().on_message(client, userdata, msg)

    def on_forecast(self, em: dict) -> None:
        """Handle weather forecast data.

        Args:
            em (dict): forecast
        """

    @override
    def run(self) -> None:
        # create, initialize and start the asynchronous thread for acquiring forecast
        self.worker = cast(
            OpenWeatherMapThread, self.instantiate(OpenWeatherMap.workerThreadId)
        )
        self.worker.init(
            self.forecast_topic,
            self.location,
            self.appid,
            self.url,
            self.update_interval,
        )
        super().run()

    @override
    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()
        data[self._OPENWEATHERMAP] = {
            "topic": self.forecast_topic,
            "location": self.location,
            "appid": self.appid,
            "url": self.url,
            "interval": self.update_interval,
        }
        return data

    @override
    def from_dict(self, data: Dict[str, Any]) -> None:
        super().from_dict(data)
        if self._OPENWEATHERMAP in data:
            for key, value in data[self._OPENWEATHERMAP].items():
                setattr(self, key, value)



from typing import Any
import unittest
from unittest.mock import patch, MagicMock
import json
from datetime import datetime, timezone
from masterpiece.mqtt import Mqtt, MqttMsg
from juham_openweathermap.openweathermap import OpenWeatherMap, OpenWeatherMapThread


class SimpleMqttMsg(MqttMsg):
    def __init__(self, topic: str, payload: Any):
        self._topic = topic
        self._payload = payload

    @property
    def payload(self) -> Any:
        return self._payload

    @payload.setter
    def payload(self, value: Any) -> None:
        self._payload = value

    @property
    def topic(self) -> str:
        return self._topic

    @topic.setter
    def topic(self, value: str) -> None:
        self._topic = value


class TestOpenWeatherMapThread(unittest.TestCase):

    @patch("juham_core.juham_cloud.requests.get")  # Mock HTTP requests
    def test_make_weburl(self, mock_get):
        thread = OpenWeatherMapThread()
        thread.init("test/topic", "Lahti,fi", "test_appid", "http://test_url", 60)

        expected_url = "http://test_url?q=Lahti,fi&APPID=test_appid"
        self.assertEqual(thread.make_weburl(), expected_url)

    @patch.object(OpenWeatherMapThread, "publish")
    def test_process_data(self, mock_publish):
        thread = OpenWeatherMapThread()
        thread.init("/test", "Lahti,fi", "test_appid", "http://test_url", 60)

        mock_response = MagicMock()
        mock_response.json.return_value = {
            "list": [
                {
                    "dt": 1710000000,
                    "main": {
                        "temp": 300,
                        "feels_like": 298,
                        "pressure": 1013,
                        "humidity": 60,
                    },
                    "wind": {"speed": 5.0},
                }
            ]
        }
        thread.process_data(mock_response)

        mock_publish.assert_called_once()
        published_msg = json.loads(mock_publish.call_args[0][1])
        self.assertEqual(published_msg[0]["temp"], 27)  # 300K - 273 = 27C
        self.assertEqual(published_msg[0]["pressure"], 1013)
        self.assertEqual(published_msg[0]["humidity"], 60)
        self.assertEqual(published_msg[0]["windspeed"], 5.0)


class TestOpenWeatherMap(unittest.TestCase):

    @patch.object(OpenWeatherMap, "subscribe")
    def test_on_connect(self, mock_subscribe):
        ow = OpenWeatherMap()
        ow.on_connect(None, None, None, 0)
        mock_subscribe.assert_called_with(ow.forecast_topic)

    @patch.object(OpenWeatherMap, "on_forecast")
    def test_on_message(self, mock_on_forecast):
        ow = OpenWeatherMap()
        test_payload = json.dumps({"temp": 25})
        msg = SimpleMqttMsg(topic=ow.forecast_topic, payload=test_payload.encode())
        ow.on_message(None, None, msg)
        mock_on_forecast.assert_called_once_with(json.loads(test_payload))

    def test_to_dict(self):
        ow = OpenWeatherMap()
        data = ow.to_dict()
        self.assertIn("OpenWeatherMap", data)
        self.assertEqual(data["OpenWeatherMap"]["location"], "Lahti,fi")

    def test_from_dict(self):
        ow = OpenWeatherMap()
        new_data = {
            "_class": "OpenWeatherMap",
            "_version": 0,
            "_object": {"name": "owm", "payload": None},
            "_base": {},
            "OpenWeatherMap": {
                "location": "NewLocation",
                "appid": "new_appid",
                "interval": 7200,
            },
        }
        ow.from_dict(new_data)
        self.assertEqual(ow.location, "NewLocation")
        self.assertEqual(ow.appid, "new_appid")
        self.assertEqual(ow.update_interval, 43200)


if __name__ == "__main__":
    unittest.main()
